import requests
import time
from .manga import Manga, MangaTag
from .chapter import Chapter
from .group import Group
from .user import User, UserSettings, UserFollow, UserUpdate
from .partial import PartialChapter, PartialGroup, PartialUser
from .relation import Relation


class APIError(Exception):
    def __init__(self, r):
        self.status = r.status_code
        self.data = None
        if not str(r.status_code).startswith("5"):
            self.data = r.json()


class LoginError(Exception):
    pass


class MangaDex:
    """Represents the MangaDex API Client."""
    def __init__(self):
        self.url = "https://mangadex.org"
        self.api = "https://api.mangadex.org/v2"
        self.session = requests.Session()
        self.login_success = False

    def login(self, username: str, password: str):
        """Logs in to MangaDex using an username and a password."""
        url = f"{self.url}/ajax/actions.ajax.php?function=login"
        credentials = {"login_username": username, "login_password": password}
        headers = {
            "method": "POST",
            "path": "/ajax/actions.ajax.php?function=login",
            "scheme": "https",
            "Accept": "*/*",
            "accept-encoding": "gzip, deflate, br",
            "Content-Type": "application/x-www-form-urlencoded; charset=UTF-8",
            "origin": self.url,
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin",
            "x-requested-with": "XMLHttpRequest",
        }

        post = self.session.post(url, data=credentials, headers=headers)
        if not post.status_code == 200:
            raise APIError(post)
        if not post.cookies.get("mangadex_session"):
            raise LoginError("Invalid credentials.")
        else:
            self.login_success = True
            return True

    def logout(self):
        """Resets the current session."""
        self.session = requests.Session()
        self.login_success = False

    def get_manga(self, id_: int, full=False) -> Manga:
        """Gets a manga with a specific id. Set full to True to populate Manga.chapters and Manga.groups"""
        p = None
        if full:
            p = {"include": "chapters"}
        req = self.session.get(f"{self.api}/manga/{id_}", params=p)

        if req.status_code == 200:
            json = req.json()["data"]
            if full:
                return Manga(json["manga"], self.session, json["chapters"], json["groups"])
            else:
                return Manga(json, self.session)
        else:
            raise APIError(req)

    def get_chapter(self, id_: int, low_quality=False, mark_read=False) -> Chapter:
        """Gets a chapter with a specific id."""
        p = {"saver": low_quality, "mark_read": mark_read}
        req = self.session.get(f"{self.api}/chapter/{id_}", params=p)

        if req.status_code == 200:
            return Chapter(req.json()["data"], self.session)
        else:
            raise APIError(req)

    def get_group(self, id_: int, full=False) -> Group:
        """Gets a group with a specific id. Set full to True to populate Group.chapters_uploaded and Group.collabs."""
        p = None
        if full:
            p = {"include": "chapters"}
        req = self.session.get(f"{self.api}/group/{id_}", params=p)

        if req.status_code == 200:
            json = req.json()["data"]
            if full:
                return Group(json["group"], json["chapters"], json["groups"])
            else:
                return Group(json)
        else:
            raise APIError(req)

    def get_user(self, id_: int = 0, full=False) -> User:
        """Gets an user with a specific id. Set full to True to populate User.chapters_uploaded and User.groups."""
        p = None
        if full:
            p = {"include": "chapters"}
        if id_ == 0 and self.login_success:
            id_ = "me"
        req = self.session.get(f"{self.api}/user/{id_}", params=p)

        if req.status_code == 200:
            json = req.json()["data"]
            if full:
                return User(json["user"], json["chapters"], json["groups"])
            else:
                return User(json)
        else:
            raise APIError(req)

    def get_user_settings(self, id_: int = 0) -> UserSettings:
        """Gets an user's settings. To retrieve another user's settings than the one currently logged in, you must
        be a MangaDex staff member."""
        if id_ == 0 and self.login_success:
            id_ = "me"
        req = self.session.get(f"{self.api}/user/{id_}/settings")

        if req.status_code == 200:
            json = req.json()["data"]
            return UserSettings(json)
        else:
            raise APIError(req)

    def get_user_list(self, id_: int = 0, follow_type: int = 0, hentai_mode: int = 1) -> []:
        """Gets an user's manga list. This settings follows the privacy mode of user's MDList."""
        p = {"hentai": hentai_mode}
        if follow_type != 0:
            p["type"] = follow_type
        if id_ == 0 and self.login_success:
            id_ = "me"
        req = self.session.get(f"{self.api}/user/{id_}/followed-manga", params=p)

        if req.status_code == 200:
            json = req.json()["data"]
            if json:
                return [UserFollow(x) for x in json]
        else:
            raise APIError(req)

    def get_user_updates(self, id_: int = 0, follow_type: int = 0, hentai_mode: int = 1, delayed=False,
                         include_blocked=False) -> []:
        """Gets an user's manga feed. To retrieve another user's feed than the one currently logged in, you must be a
        MangaDex staff member."""
        p = {"type": follow_type, "hentai": hentai_mode, "delayed": delayed, "blockgroups": include_blocked}
        if id_ == 0 and self.login_success:
            id_ = "me"
        req = self.session.get(f"{self.api}/user/{id_}/followed-updates", params=p)

        if req.status_code == 200:
            json = req.json()["data"]["chapters"]
            if json:
                return [UserUpdate(x) for x in json]
        else:
            raise APIError(req)

    def get_user_ratings(self, id_: int = 0) -> {}:
        """Gets an user's ratings."""
        if id_ == 0 and self.login_success:
            id_ = "me"
        req = self.session.get(f"{self.api}/user/{id_}/ratings")

        if req.status_code == 200:
            json = req.json()["data"]
            if json:
                return {x["mangaId"]: x["rating"] for x in json}
        else:
            raise APIError(req)

    def get_user_manga(self, id_: int, uid: int = 0) -> UserFollow:
        """Gets an user's manga from their MDList."""
        if uid == 0 and self.login_success:
            uid = "me"
        req = self.session.get(f"{self.api}/user/{uid}/manga/{id_}")

        if req.status_code == 200:
            json = req.json()["data"]
            return UserFollow(json)
        else:
            raise APIError(req)

    def set_user_markers(self, mangas: list, read: bool, id_: int = 0):
        """Sets chapters as read or unread."""
        reqs = []
        lists = [mangas[x:x + 100] for x in range(0, len(mangas), 100)]
        if id_ == 0 and self.login_success:
            id_ = "me"

        for y in lists:
            p = {"chapters": y, "read": read}
            reqs.append(self.session.post(f"{self.api}/user/{id_}/marker", json=p))
            time.sleep(1)

        if reqs[-1].status_code == 200:
            return True
        else:
            raise APIError(reqs[-1])



from .partial import PartialChapter, PartialGroup
hentai_mode = {0: "disabled", 1: "show", 2: "only"}


class User:
    """Represents a MangaDex User."""
    __slots__ = ("id", "username", "level_id", "joined", "last_seen", "website", "bio", "views", "uploads", "premium",
                 "md_ah", "avatar", "chapters_uploaded", "groups")

    def __init__(self, data, chaps=None, groups=None):
        self.id = data["id"]
        self.username = data["username"]
        self.level_id = data["levelId"]
        self.joined = data["joined"]
        self.last_seen = data["lastSeen"]
        self.website = data["website"]
        self.bio = data["biography"]
        self.views = data["views"]
        self.uploads = data["uploads"]
        self.premium = data["premium"]
        self.md_ah = data["mdAtHome"]
        self.avatar = data["avatar"]
        self.chapters_uploaded = [PartialChapter(x) for x in chaps or []]
        self.groups = [PartialGroup(x) for x in groups or []]


class UserSettings:
    """Represents a MangaDex User's settings."""
    __slots__ = ("id", "hentai_mode", "latest_update", "show_moderated_posts", "show_unavailable_chapters",
                 "whitelisted_languages", "blacklisted_tags")

    def __init__(self, data):
        self.id = data["id"]
        self.hentai_mode = hentai_mode[data["hentaiMode"]]
        self.latest_update = data["latestUpdates"]
        self.show_moderated_posts = data["showModeratedPosts"]
        self.show_unavailable_chapters = data["showUnavailableChapters"]
        self.whitelisted_languages = data["shownChapterLangs"]
        self.blacklisted_tags = data["excludedTags"]


class UserFollow:
    """Represents a MangaDex User's followed manga."""
    __slots__ = ("user", "id", "title", "hentai", "follow_type", "user_volume", "user_chapter", "user_rating", "cover")

    def __init__(self, data):
        self.user = data["userId"]
        self.id = data["mangaId"]
        self.title = data["mangaTitle"]
        self.hentai = data["isHentai"]
        self.follow_type = data["followType"]
        self.user_volume = data["volume"]
        self.user_chapter = data["chapter"]
        self.user_rating = data["rating"]
        self.cover = data["mainCover"]


class UserUpdate:
    """Represents a MangaDex User's feed item."""
    __slots__ = ("id", "hash", "manga_id", "manga_title", "volume", "chapter", "title", "language", "groups",
                 "uploader", "timestamp", "thread_id", "comments", "views", "read")

    def __init__(self, data):
        self.id = data["id"]
        self.hash = data["hash"]
        self.manga_id = data["mangaId"]
        self.manga_title = data["mangaTitle"]
        self.volume = data["volume"]
        self.chapter = data["chapter"]
        self.title = data["title"]
        self.language = data["language"]
        self.groups = data["groups"]
        self.uploader = data["uploader"]
        self.timestamp = data["timestamp"]
        self.thread_id = data["threadId"]
        self.comments = data["comments"]
        self.views = data["views"]
        self.read = data["read"]



class PartialChapter:
    """Represents a MangaDex Chapter, but with reduced info."""
    __slots__ = ("id", "hash", "manga_id", "manga_title", "volume", "chapter", "title", "language", "groups",
                 "uploader", "timestamp", "thread_id", "views")

    def __init__(self, data):
        self.id = data["id"]
        self.hash = data["hash"]
        self.manga_id = data["mangaId"]
        self.manga_title = data["mangaTitle"]
        self.volume = data["volume"]
        self.chapter = data["chapter"]
        self.title = data["title"]
        self.language = data["language"]
        self.groups = data["groups"]
        self.uploader = data["uploader"]
        self.timestamp = data["timestamp"]
        self.thread_id = data["threadId"]
        self.views = data["views"]


class PartialGroup:
    """Represents a MangaDex Group, but with reduced info."""
    __slots__ = ("id", "name")

    def __init__(self, data):
        self.id = data["id"]
        self.name = data["name"]


class PartialUser:
    """Represents a MangaDex User, but with reduced info."""
    __slots__ = ("id", "username")

    def __init__(self, data):
        self.id = data["id"]
        self.username = data["name"]



relations = {1: "prequel", 2: "sequel", 3: "adaptation", 4: "spin_off", 5: "side_story", 6: "main", 7: "alternate",
             8: "doujinshi", 9: "based", 10: "colored", 11: "monochrome", 12: "shared_universe", 13: "same_franchise",
             14: "pre_serialization", 15: "serialization"}


class Relation:
    """Represents a MangaDex Manga Relation."""
    __slots__ = ("id", "title", "type", "hentai")

    def __init__(self, data):
        self.id = data["id"]
        self.title = data["title"]
        self.type = relations[data["type"]]
        self.hentai = data["isHentai"]
